require "GUSI.ph";

$macUsage=<<EofMacUsage;
macweblint 97 v$VERSION - 
   pick fluff off web pages (HTML)

Drag and drop file or folder icons onto `macweblint'!!\n\n
EofMacUsage
$macUnchange = "Preferences are not altered.\n";
@macOpts = ('-i', '-pedantic', '-s', '-t');
@macOptMeans = (
	'ignore case in element tags',
	'turn on all warnings, except for case of element tags',
	'give short warning messages (filename not printed)',
	'terse warning mode, useful mainly for testsuite'
);
@macExtMeans = (
	'HTML extension to include (HTML3.2)',
	'HTML extension to include (i18n)',
	'HTML extension to include (Cougar)',
	'HTML extension to include (Microsoft)',
	'HTML extension to include (Netscape)',
	'no extension (refer only to DOCTYPE) (-) [default]'
);
@macCodeMeans = (
	'specify the code type (literal) [default]',
	'specify the code type (ISO-2022-7bit)',
	'specify the code type (Unicode)'
);

sub macGetOpts {
	local($option, $output, @options);
	if (open(OPT, $MAC_OPTFILE)) {
		$output = scalar(<OPT>);
		chop $output;
		@options = split(/\s+/, <OPT>);
		close(OPT);
		$option = join(" ", @options);
	} else {
		open(OPT, ">$MAC_OPTFILE") || die "$PROGRAM: file ".
			"Can't creat a file `$MAC_OPTFILE' !\n";
		print OPT "MacPerl console\n\n";
		close(OPT);
		print "$PROGRAM: The file `$MAC_OPTFILE' is not found,
			so it was created as a new file!\n";
		$output = 'MacPerl console';
		$option = '';
	}
	return ($output, $option);
}

sub macPutOpts {
	local($output, $option) = @_;
	open(OPT, ">$MAC_OPTFILE") || 
		die "$PROGRAM: The file `$MAC_OPTFILE' is not created!\n";
	print OPT "$output\n$option\n";
	close(OPT);
	exit 0;
}

sub macUsage {
	local($worknum, $newopt);

	$worknum = &MacPerl'Answer($macUsage, 
		"Quit", "options", "output");
	if ($worknum == 0) {
		&macChangeOutput;
	} elsif ($worknum == 1) {
		&macChangeOpt;
	} else {
		exit 0;
	}
}

sub macChangeOpt {
	local($newopt, $way, $yesno);

	$way = &MacPerl'Answer("Change option\n".
		" (current option: $macOption)", "Choose", "Type", "Cancel");
	if ($way == 2) {
		$newopt = &macSelectOpt($macOption);
	} elsif ($way == 1) {
		$newopt = &macInputOpt($macOption);
	} else {
		die $macUnchange;
	}
	die $macUnchange if ($newopt eq $macOption);
	$yesno = &MacPerl'Answer("confirmation of changing options.\n".
		"  (formerly: $macOption)\n".
		"  (current: $newopt)", "Ok", "Cancel");
	die $macUnchange unless ($yesno);
	print "$PROGRAM: The options are changed as `${newopt}'.\n".
		"(formerly: `${macOption}'\n";
	&macPutOpts($macOutput, $newopt);
	exit 0;
}

sub macInputOpt {
	local($newopt);

	$newopt = &MacPerl'Ask("Please type new options.",
		$macOption);
	die $macUnchange unless defined $newopt;
	$newopt;
}

sub macSelectOpt {
	local($opt, @newopts, $newext);
	local($nowopt) = "$macOption ";

	($nowopt, $newext) = &macSelectExtension($nowopt);
	push(@newopts, "-x $newext") if ($newext ne '-');
	($nowopt, $newcode) = &macSelectCode($nowopt);
	push(@newopts, "-c $newcode");
	foreach $opt (@macOpts) {
		local($optmean) = shift @macOptMeans;
		local($onoff, $onoffmes, $way);
		$onoff = ($nowopt =~ s/$opt //i);
		$onoffmes = $onoff ? "Yes" : "No";
		$way = &MacPerl'Answer("$optmean\n".
			"  (current: $onoffmes)", "Unchange",
			"Change", "Cancel");
		die $macUnchange if ($way == 0);
		$onoff = !$onoff if ($way == 1);
		push(@newopts, $opt) if ($onoff);
	}
	if ($nowopt ne '') {
		chop $nowopt;
		push(@newopts, $nowopt);
	}
	return join(' ', @newopts);
}

sub macSelectExtension {
	local($nowopt) = @_;
	$nowopt =~ s/-x ([\w\d.-]+) //i;
	$nowext = ($1 eq '') ? '-' : $1;
	local($value) = &MacPerl'Pick("Please select HTML extension".
		" to include. (current `$nowext')", @macExtMeans);
	local($ext) = ($value =~ /\(([\w\d.-]+)\)/);
	die $macUnchange if ($ext eq '');
	return ($nowopt, $ext);
}

sub macSelectCode {
	local($nowopt) = @_;
	$nowopt =~ s/-c ([\w\d.-]+) //i;
	$nowext = ($1 eq '') ? 'literal' : $1;
	local($value) = &MacPerl'Pick("Please select the code type.".
		"(current `$nowext')", @macCodeMeans);
	local($ext) = ($value =~ /\(([\w\d.-]+)\)/);
	die $macUnchange if ($ext eq '');
	return ($nowopt, $ext);
}

sub macChangeOutput {
	local($way, $newoutput);

	$way = &MacPerl'Answer("Change output\n  (current: $macOutput)",
		"Choose file", "console", "Cancel"), "\n";
	if ($way == 2) {
		$newoutput = &MacPerl'Choose(&GUSI'AF_FILE, 0, "", "",
			&GUSI'CHOOSE_NEW + &GUSI'CHOOSE_DEFAULT,
			"jweblint.log");
		die $macUnchange unless defined $newoutput;
	} elsif ($way == 1) {
		$newoutput = 'MacPerl console';
	} else {
		die $macUnchange if ($way == 0);
	}
	die $macUnchange if ($newoutput eq $macOutput);
	$yesno = &MacPerl'Answer("confirmation of changing options\n".
		"  (formerly: $macOutput)\n".
		"  (current: $newoutput)", "Ok", "Cancel");
	die $macUnchange unless ($yesno);
	print "$PROGRAM: The result will output to `${newoutput}'.".
		"(formerly: `${macOutput}')\n",
	&macPutOpts($newoutput, $macOption);
}

1;
